import datetime
import enum
from dataclasses import dataclass
from decimal import Decimal
from typing import List, Dict, Tuple, Optional
import logging

import pandas as pd
from eth_defi.abi import get_contract
from eth_defi.event_reader.conversion import decode_data, convert_int256_bytes_to_int
from eth_defi.event_reader.logresult import LogResult, LogContext
from eth_defi.event_reader.reader import read_events_concurrent
from eth_defi.event_reader.web3factory import TunedWeb3Factory
from eth_defi.event_reader.web3worker import create_thread_pool_executor
from eth_defi.price_oracle.oracle import PriceOracle
from eth_defi.uniswap_v2.pair import PairDetails

from .trade_feed import OHLCVProducer, Trade, TradeDelta
from .reorg_mon import ReorganisationMonitor

logger = logging.getLogger(__name__)


#: List of output columns to pairs.csv
PAIR_FIELD_NAMES = [
    "block_number",
    "timestamp",
    "tx_hash",
    "log_index",
    "factory_contract_address",
    "pair_contract_address",
    "pair_count_index",
    "token0_address",
    "token0_symbol",
    "token1_address",
    "token1_symbol",
]

#: List of fields we need to decode in swaps
SWAP_FIELD_NAMES = [
    "block_number",
    "timestamp",
    "tx_hash",
    "log_index",
    "pair_contract_address",
    "amount0_in",
    "amount1_in",
    "amount0_out",
    "amount1_out",
]

#: List of fields we need to decode in syncs
SYNC_FIELD_NAMES = [
    "block_number",
    "timestamp",
    "tx_hash",
    "log_index",
    "pair_contract_address",
    "reserve0",
    "reserve1",
]


class SwapKind(enum.Enum):
    """What kind of swaps we might have."""

    # token1 -> token0
    buy = "buy"

    # token0 -> token1
    sell = "sell"

    # Traded both ways at the same time
    complex = "complex"

    # Zero traded volumne
    invalid = "invalid"


class UniswapV2OHLCVProducer(OHLCVProducer):
    """Uniswap v2 compatible DXE candle generator.

    Uses multiple threads to speed up blockchain read.
    """

    def __init__(self,
                 pairs: List[PairDetails],
                 web3_factory: TunedWeb3Factory,
                 oracles: Dict[str, PriceOracle],
                 reorg_mon: ReorganisationMonitor,
                 data_retention_time: Optional[pd.Timedelta] = None,
                 candle_size=pd.Timedelta(minutes=1),
                 threads=16,
                 chunk_size=100):

        super().__init__(
            oracles=oracles,
            reorg_mon=reorg_mon,
            data_retention_time=data_retention_time,
            candle_size=candle_size,
        )

        #: Pair address -> details mapping
        self.pair_map = Dict[str, PairDetails] = {p.address: p for p in pairs}
        self.web3_factory = web3_factory
        self.event_reader_context = LogContext()
        self.chunk_size = chunk_size
        self.executor = create_thread_pool_executor(self.web3_factory, self.event_reader_context, max_workers=threads)

        # A web3 instance used in the main thread
        self.web3 = web3_factory(self.event_reader_context)

        # Get data from ABI
        Pair = get_contract(self.web3, "UniswapV2Pair.json")
        self.events_to_read = [Pair.events.Swap, Pair.events.Sync]

    def get_block_number_at_chain_tip(self) -> int:
        return self.web3.eth.block_number

    def update_block_range(self, start_block: int, end_block: Optional[int]) -> TradeDelte:
        """Read data between logs.

        :raise ChainReorganisationDetected:
            In the case we notice chain data has changed during the reading
        """

        events = []

        def _extract_timestamps(web3, start_block, end_block):
            ts_map = {}
            for block_num in range(start_block, end_block+1):
                ts_map[block_num] = self.reorg_mon.get_block_timestamp(block_num)
            return ts_map

        # Read specified events in block range
        for log_result in read_events_concurrent(
            self.executor,
            start_block,
            end_block,
            self.events_to_read,
            None,
            chunk_size=self.chunk_size,
            context=self.event_reader_context,
            extract_timestamps=_extract_timestamps,
        ):
            if log_result["event"].event_name == "Swap":
                events.append(decode_swap(log_result))
            elif log_result["event"].event_name == "Sync":
                events.append(decode_swap(log_result))

        trades = self.map_uniswap_v2_events(events)
        self.add_trades(trades)

    def map_uniswap_v2_events(self, events: List[dict]) -> List[Trade]:
        """Figure out Uniswap v2 swap and volume."""

        prev_sync = None
        trades = []
        for evt in events:
            if evt["type"] == "sync":
                native_price = Decimal(evt["reserve0"]) / Decimal(evt["reserve1"])
                prev_sync = evt
            else:
                # Swap
                tx_hash = evt["tx_hash"]
                if prev_sync["tx_hash"] != tx_hash:
                    logger.debug("Current sync and swap do not follow Uniswap logic: %s - %s", prev_sync, evt)

                pair: PairDetails
                pair = self.pair_map[evt["pair_contract_address"]]

                price, amount = calculate_reserve_price_in_quote_token_raw(
                    pair.reverse_token_order,
                    prev_sync["reserve0"],
                    prev_sync["reserve1"],
                    evt["amount0_in"],
                    evt["amount1_in"],
                    evt["amount0_out"],
                    evt["amount1_out"],
                )

                timestamp = self.reorg_mon.get_block_timestamp(evt["block_number"])

                t = Trade(
                    pair=pair.address,
                    block_number=evt["block_number"],
                    block_hash=evt["block_hash"],
                    log_index=evt["log_index"],
                    tx_hash=evt["tx_hash"],
                    timestamp=pd.Timestamp.utcfromtimestamp(timestamp),
                    price=price,
                    amount=amount,
                )
                trades.append(t)
        return trades


def decode_swap(log: LogResult) -> dict:
    """Process swap event.

    This function does manually optimised high speed decoding of the event.

    The event signature is:

    .. code-block::

        event Swap(
          address indexed sender,
          uint amount0In,
          uint amount1In,
          uint amount0Out,
          uint amount1Out,
          address indexed to
        );
    """

    # Raw example event
    # {'address': '0xb4e16d0168e52d35cacd2c6185b44281ec28c9dc', 'blockHash': '0x4ba33a650f9e3d8430f94b61a382e60490ec7a06c2f4441ecf225858ec748b78', 'blockNumber': '0x98b7f6', 'data': '0x00000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000046ec814a2e900000000000000000000000000000000000000000000000000000000000003e80000000000000000000000000000000000000000000000000000000000000000', 'logIndex': '0x4', 'removed': False, 'topics': ['0xd78ad95fa46c994b6551d0da85fc275fe613ce37657fb8d5e3d130840159d822', '0x000000000000000000000000f164fc0ec4e93095b804a4795bbe1e041497b92a', '0x0000000000000000000000008688a84fcfd84d8f78020d0fc0b35987cc58911f'], 'transactionHash': '0x932cb88306450d481a0e43365a3ed832625b68f036e9887684ef6da594891366', 'transactionIndex': '0x1', 'context': <__main__.TokenCache object at 0x104ab7e20>, 'event': <class 'web3._utils.datatypes.Swap'>, 'timestamp': 1588712972}

    block_time = datetime.datetime.utcfromtimestamp(log["timestamp"]).tz_localize(None)

    pair_contract_address = log["address"]

    # Chop data blob to byte32 entries
    data_entries = decode_data(log["data"])

    amount0_in, amount1_in, amount0_out, amount1_out = data_entries

    data = {
        "type": "sync",
        "block_number": int(log["blockNumber"], 16),
        "timestamp": block_time.isoformat(),
        "tx_hash": log["transactionHash"],
        "log_index": int(log["logIndex"], 16),
        "pair_contract_address": pair_contract_address,
        "amount0_in": convert_int256_bytes_to_int(amount0_in),
        "amount1_in": convert_int256_bytes_to_int(amount1_in),
        "amount0_out": convert_int256_bytes_to_int(amount0_out),
        "amount1_out": convert_int256_bytes_to_int(amount1_out),
    }
    return data


def decode_sync(log: LogResult) -> dict:
    """Process sync event.

    This function does manually optimised high speed decoding of the event.

    The event signature is:

    .. code-block::

        event Sync(uint112 reserve0, uint112 reserve1);
    """

    # Raw example event
    # {'address': '0xb4e16d0168e52d35cacd2c6185b44281ec28c9dc', 'blockHash': '0x4ba33a650f9e3d8430f94b61a382e60490ec7a06c2f4441ecf225858ec748b78', 'blockNumber': '0x98b7f6', 'data': '0x00000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000046ec814a2e900000000000000000000000000000000000000000000000000000000000003e80000000000000000000000000000000000000000000000000000000000000000', 'logIndex': '0x4', 'removed': False, 'topics': ['0xd78ad95fa46c994b6551d0da85fc275fe613ce37657fb8d5e3d130840159d822', '0x000000000000000000000000f164fc0ec4e93095b804a4795bbe1e041497b92a', '0x0000000000000000000000008688a84fcfd84d8f78020d0fc0b35987cc58911f'], 'transactionHash': '0x932cb88306450d481a0e43365a3ed832625b68f036e9887684ef6da594891366', 'transactionIndex': '0x1', 'context': <__main__.TokenCache object at 0x104ab7e20>, 'event': <class 'web3._utils.datatypes.Swap'>, 'timestamp': 1588712972}

    block_time = datetime.datetime.utcfromtimestamp(log["timestamp"])

    pair_contract_address = log["address"]

    # Chop data blob to byte32 entries
    data_entries = decode_data(log["data"])

    reserve0, reserve1 = data_entries

    data = {
        "type": "sync",
        "block_number": int(log["blockNumber"], 16),
        "timestamp": block_time.isoformat(),
        "tx_hash": log["transactionHash"],
        "log_index": int(log["logIndex"], 16),
        "pair_contract_address": pair_contract_address,
        "reserve0": convert_int256_bytes_to_int(reserve0),
        "reserve1": convert_int256_bytes_to_int(reserve1),
    }
    return data


def calculate_reserve_price_in_quote_token_raw(
        reversed: bool,
        reserve0: int,
        reserve1: int,
        amount0_in,
        amount1_in,
        amount0_out,
        amount1_out,
) -> Tuple[Decimal, Decimal]:
    """Calculate the market price based on Uniswap pool reserve0 an reserve1.

    :param reversed:
        Determine base, quote token order relative to token0, token1.
        If reversed, quote token is token0, else quote token is token0.

    :return:
        Price in quote token, amount in quote token
    """

    assert reserve0 > 0, f"Bad reserves {reserve0}, {reserve1}"
    assert reserve1 > 0, f"Bad reserves {reserve0}, {reserve1}"

    if reversed:
        reserve0, reserve1 = reserve1, reserve0

    if reversed:
        quote_amount = (amount0_out - amount0_in)
        base_amount = (amount1_out - amount1_in)
    else:
        base_amount = (amount0_out - amount0_in)
        quote_amount = (amount1_out - amount1_in)

    price = Decimal(reserve1) / Decimal(reserve0)

    volume = Decimal(quote_amount)

    return price, volume







"""Timeframe helpers."""

from dataclasses import dataclass

import pandas as pd


@dataclass
class Timeframe:
    """Describe candle timeframe.

    This structure allows us to pass candle resample
    data around the framework.
    """

    #: Pandas frequency string.
    #:
    #: E.g. `1D` for daily, `1min` for minute.
    #:
    #: See https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#offset-aliases
    freq: str

    #: Candle shift offset
    #:
    #: E.g. move hourly candles 5 minutes backwards to start at 00:55
    offset: pd.Timedelta = pd.Timedelta(seconds=0)

    def round_timestamp_down(self, ts: pd.Timestamp) -> pd.Timestamp:
        """Snap to previous available timedelta.

        Preserve any timezone info on `ts`.

        If we are at the the given exact delta, then do not round, only add offset.

        :param ts:
            Timestamp we want to round

        :return:
            When to wake up from the sleep next time
        """
        return ts.floor(self.freq) + self.offset



"""Chain reorganisation handling during the real-time OHLCV candle production."""

import datetime
import time
from abc import abstractmethod
from dataclasses import dataclass
from typing import Dict, Iterable, Optional, Tuple
import logging

import pandas as pd
from web3 import Web3

logger = logging.getLogger(__name__)


@dataclass(slots=True, frozen=True)
class BlockRecord:
    block_number: int
    block_hash: str
    timestamp: float | int

    def __post_init__(self):
        assert type(self.block_number) == int
        assert type(self.block_hash) == str
        assert type(self.timestamp) in (float, int)


@dataclass(slots=True, frozen=True)
class ChainReorganisationResolution:

    #: What we know is the chain tip on our node
    last_live_block: int

    #: What we know is the block for which we do not need to perform rollback
    latest_block_with_good_data: int

    #: Did we detect any reorgs in this chycle
    reorg_detected: bool


class ChainReorganisationDetected(Exception):
    block_number: int
    original_hash: str
    new_hash: str

    def __init__(self, block_number: int, original_hash: str, new_hash: str):
        self.block_number = block_number
        self.original_hash = original_hash
        self.new_hash = new_hash

        super().__init__(f"Block reorg detected at #{block_number:,}. Original hash: {original_hash}. New hash: {new_hash}")


class ReorganisationResolutionFailure(Exception):
    """Chould not figure out chain reorgs after mutliple attempt.

    Node in a bad state?
    """


class BlockNotAvailable(Exception):
    """Tried to ask timestamp data for a block that does not exist yet."""



class ReorganisationMonitor:
    """Watch blockchain for reorgs.

    - Maintain the state of the last read block

    - Check block headers for chain reorganisations

    - Also manages the service for block timestamp lookups
    """

    def __init__(self, check_depth=200, max_reorg_resolution_attempts=10, reorg_wait_seconds=5):
        self.block_map: Dict[int, BlockRecord] = {}
        self.last_block_read: int = 0
        self.check_depth = check_depth
        self.reorg_wait_seconds = reorg_wait_seconds
        self.max_cycle_tries = max_reorg_resolution_attempts

    def has_data(self) -> bool:
        """Do we have any data available yet."""
        return len(self.block_map) > 0

    def get_last_block_read(self):
        return self.last_block_read

    def load_initial_data(self, block_count: int) -> Tuple[int, int]:
        """Get the inital block buffer filled up.

        :return:
            The initial block range to start to work with
        """
        end_block = self.get_last_block_live()
        start_block = max(end_block - block_count, 1)

        for block in self.get_block_data(start_block, end_block):
            self.add_block(block)

        return start_block, end_block

    def add_block(self, record: BlockRecord):
        """Add new block to header tracking.

        Blocks must be added in order.
        """
        block_number = record.block_number
        assert block_number not in self.block_map, f"Block already added: {block_number}"
        self.block_map[block_number] = record

        assert self.last_block_read == block_number - 1, f"Blocks must be added in order. Last: {self.last_block_read}, got: {record}"
        self.last_block_read = block_number

    def check_block_reorg(self, block_number: int, block_hash: str):
        """Check that newly read block matches our record.

        If we do not have record, ignore.
        """
        original_block = self.block_map.get(block_number)
        if original_block is not None:
            if original_block.block_hash != block_hash:
                raise ChainReorganisationDetected(block_number, original_block.block_hash, block_hash)

    def truncate(self, latest_good_block: int):
        """Delete data after a block number because chain reorg happened.

        :param latest_good_block:
            Delete all data starting after this block (exclusive)
        """
        assert self.last_block_read
        for block_to_delete in range(latest_good_block + 1, self.last_block_read + 1):
            del self.block_map[block_to_delete]
        self.last_block_read = latest_good_block

    def figure_reorganisation_and_new_blocks(self):
        """Compare the local block database against the live data from chain.

        Spot the differences in (block number, block header) tuples
        and determine a chain reorg.
        """
        chain_last_block = self.get_last_block_live()
        check_start_at = max(self.last_block_read - self.check_depth, 1)
        for block in self.get_block_data(check_start_at, chain_last_block):
            self.check_block_reorg(block.block_number, block.block_hash)
            if block.block_number not in self.block_map:
                self.add_block(block)

    def get_block_timestamp(self, block_number: int) -> int:
        """Return UNIX UTC timestamp of a block."""

        if not self.block_map:
            raise BlockNotAvailable("We have no records of any blocks")

        if block_number not in self.block_map:
            last_recorded_block_num = max(self.block_map.keys())
            raise BlockNotAvailable(f"Block {block_number} has not data, the latest live block is {self.get_last_block_live()}, last recorded is {last_recorded_block_num}")

        return self.block_map[block_number].timestamp

    def get_block_timestamp_as_pandas(self, block_number: int) -> pd.Timestamp:
        """Return UNIX UTC timestamp of a block."""

        ts = self.get_block_timestamp(block_number)
        return pd.Timestamp.utcfromtimestamp(ts).tz_localize(None)

    def update_chain(self) -> ChainReorganisationResolution:
        """Attemp

        - Do several attempt to read data (as a fork can cause other forks can cause fork)

        - Give up after some time if we detect the chain to be in a doom loop

        :return:
            What we think about the chain state
        """

        tries_left = self.max_cycle_tries
        max_purge = self.get_last_block_read()
        reorg_detected = False
        while tries_left > 0:
            try:
                self.figure_reorganisation_and_new_blocks()
                return ChainReorganisationResolution(self.last_block_read, max_purge, reorg_detected=reorg_detected)
            except ChainReorganisationDetected as e:
                logger.info("Chain reorganisation detected: %s", e)

                latest_good_block = e.block_number - 1

                reorg_detected = True

                if max_purge:
                    max_purge = min(latest_good_block, max_purge)
                else:
                    max_purge = e.block_number

                self.truncate(latest_good_block)
                tries_left -= 1
                time.sleep(self.reorg_wait_seconds)

        raise ReorganisationResolutionFailure(f"Gave up chain reorg resolution. Last block: {self.last_block_read}, attempts {self.max_cycle_tries}")

    @abstractmethod
    def get_block_data(self, start_block, end_block) -> Iterable[BlockRecord]:
        """Read the new block headers.

        :param start_block:
            The first block where to read (inclusive)

        :param end_block:
            The block where to read (inclusive)
        """

    @abstractmethod
    def get_last_block_live(self) -> int:
        """Get last block number"""


class JSONRPCReorganisationMonitor(ReorganisationMonitor):
    """Watch blockchain for reorgs using eth_getBlockByNumber JSON-RPC API."""

    def __init__(self, web3: Web3, check_depth=200, max_reorg_resolution_attempts=10):
        super().__init__(check_depth=check_depth, max_reorg_resolution_attempts=max_reorg_resolution_attempts)
        self.web3 = web3

    def get_last_block_live(self):
        return self.web3.eth.block_number

    def get_block_data(self, start_block, end_block) -> Iterable[BlockRecord]:
        logger.debug("Extracting timestamps for logs %d - %d", start_block, end_block)
        web3 = self.web3

        # Collect block timestamps from the headers
        for block_num in range(start_block, end_block + 1):
            raw_result = web3.manager.request_blocking("eth_getBlockByNumber", (hex(block_num), False))
            data_block_number = raw_result["number"]
            block_hash = raw_result["hash"]
            assert type(data_block_number) == str, "Some automatic data conversion occured from JSON-RPC data. Make sure that you have cleared middleware onion for web3"
            assert int(raw_result["number"], 16) == block_num

            timestamp = int(raw_result["timestamp"], 16)

            yield BlockRecord(block_num, block_hash, timestamp)


class MockChainAndReorganisationMonitor(ReorganisationMonitor):
    """A dummy reorganisation monitor for unit testing.

    Simulate block production and chain reorgs by minor forks,
    like a real blockchain.
    """

    def __init__(self, block_number: int = 1, block_duration_seconds=1):
        super().__init__(reorg_wait_seconds=0)
        self.simulated_block_number = block_number
        self.simulated_blocks = {}
        self.block_duration_seconds = block_duration_seconds

    def produce_blocks(self, block_count=1):
        """Populate the fake blocks in mock chain.

        These blocks will be "read" in py:meth:`figure_reorganisation_and_new_blocks`.
        """
        for x in range(block_count):
            num = self.simulated_block_number
            record = BlockRecord(num, hex(num), num * self.block_duration_seconds)
            self.simulated_blocks[self.simulated_block_number] = record
            self.simulated_block_number += 1

    def produce_fork(self, block_number: int, fork_marker="0x8888"):
        """Mock a fork int he chain."""
        self.simulated_blocks[block_number] = BlockRecord(block_number, fork_marker, block_number * self.block_duration_seconds)

    def get_last_block_live(self):
        return self.simulated_block_number - 1

    def get_block_data(self, start_block, end_block) -> Iterable[BlockRecord]:

        assert start_block > 0, "Cannot ask data for zero block"
        assert end_block <= self.get_last_block_live(), "Cannot ask data for blocks that are not produced yet"

        for i in range(start_block, end_block + 1):
            yield self.simulated_blocks[i]





from dataclasses import dataclass
from typing import List

import pandas as pd

from tradingstrategy.direct_feed.ohlcv_aggregate import resample_trades_into_ohlcv, get_feed_for_pair, truncate_ohlcv
from tradingstrategy.direct_feed.timeframe import Timeframe
from tradingstrategy.direct_feed.trade_feed import TradeDelta
from tradingstrategy.direct_feed.direct_feed_pair import PairId


class CandleFeed:
    """Create candles for certain time frame for multiple pairs.

    - Takes :py:class:`TradeFeed` as input

    - Generates candles based on this feed

    - Can only generate candles of one duration
    """

    def __init__(self,
                 pairs: List[PairId],
                 timeframe: Timeframe,
                 ):
        """

        :param freq:
            Pandas frequency string e.g. "1H", "min"

        :param candle_offset:
        """
        self.timeframe = timeframe
        self.candle_df = pd.DataFrame()
        self.last_cycle = 0

    def apply_delta(self, delta: TradeDelta):
        """Add new candle data generated from the latest blockchain input."""
        cropped_df = truncate_ohlcv(self.candle_df, delta.start_ts)
        candles = resample_trades_into_ohlcv(delta.trades, self.timeframe)
        self.candle_df = pd.concat([cropped_df, candles])
        self.last_cycle = delta.cycle

    def get_candles_by_pair(self, pair: PairId) -> pd.DataFrame:
        return get_feed_for_pair(self.candle_df, pair)

    def get_last_block_number(self) -> int:
        """Get overall last block number for which we have valid data.

        :return:
            block number (inclusive)
        """
        return self.candle_df["end_block"].max()









from abc import abstractmethod
from dataclasses import dataclass, asdict
from decimal import Decimal
from typing import Dict, Optional, List,  Iterable, Type, TypeAlias

import pandas as pd
from eth_defi.price_oracle.oracle import BaseOracle

from tqdm import tqdm

from .direct_feed_pair import PairId
from .reorg_mon import ReorganisationMonitor, ChainReorganisationResolution
from .timeframe import Timeframe


#: Hex presentation of transaction hash, such
Binary256: TypeAlias = str


@dataclass(slots=True, frozen=True)
class Trade:
    """Capture information about single trade.

    Designed for technical analysis and trading,
    prices are not intentionally unit accurate and thus
    not suitable for accounting.
    """

    #: Trading pair od
    pair: PairId

    #: Block number
    block_number: int

    #: Block hash
    block_hash: str

    #: Block mined at
    timestamp: pd.Timestamp

    #: Block mined at
    tx_hash: Binary256

    #: Trade index within the block
    log_index: int

    #: Trade price in quote token
    price: Decimal

    #: Trade amount in quote token.
    #:
    #: Positive for buys, negative for sells.
    amount: Decimal

    #: The US dollar conversion rate used for this rate
    exchange_rate: Decimal

    def __post_init__(self):
        assert self.timestamp.tzinfo is None, "Don't use timezone aware timestamps - everything must be naive UTC"
        assert self.tx_hash, "tx_hash missing"

    @staticmethod
    def get_dataframe_columns() -> dict:
        fields = dict([
            ("pair", "string"),
            ("block_number", "uint64"),
            ("block_hash", "string"),
            ("timestamp", "datetime64[s]"),
            ("tx_hash", "string"),
            ("log_index", "uint32"),
            ("price", "object"),
            ("amount", "object"),
            ("exchange_rate", "object"),
        ])
        return fields


@dataclass(slots=True, frozen=True)
class TradeDelta:
    """Trade history delta for a snapshot"""

    #: Running counter for each connected client
    cycle: int

    #: New incoming data starting this block.
    #:
    #: The client might have already data older than > start_block
    #: from previous delta. In this case the client needs to
    #: purhe  this data as it is discarded due to chain reorganisation.
    start_block: int

    # The start block we are using if we do not do timeframe snap.
    #
    unadjusted_start_block: int

    #: Last block for which we have data
    #:
    #:
    end_block: int

    #: Timestamp of the start block
    #:
    start_ts: pd.Timestamp

    #: Timestamp of the end block
    #:
    end_ts: pd.Timestamp

    #: Did we detect any chain reorgs on this update cycle?
    #:
    reorg_detected: bool

    #: List of individual trades
    #:
    trades: pd.DataFrame


class TradeFeed:
    """Real-time blockchain trade feed.

    In-memory latency optimised OHLCv producer for on-chain trades.

    - Keep events in RAM

    - Generate candles based on events

    - Gracefully handle chain reorganisations and drop
      stale data
    """

    def __init__(self,
                 pairs: List[PairId],
                 oracles: Dict[PairId, BaseOracle],
                 reorg_mon: ReorganisationMonitor,
                 timeframe: Timeframe = Timeframe("1min"),
                 data_retention_time: Optional[pd.Timedelta] = None,
                 ):
        """
        Create new real-time OHLCV tracker.

        :param pairs:
            List of pool addresses, or list of similar identifiers.

        :param oracles:
            Reference prices for converting ETH or other crypto quoted
            prices to US dollars.

            In the form of pair -> Price oracle maps.

        :param reorg_mon:
            Reorganisation detector and last good block state for a chain.

        :param timeframe:
            Expected timeframe of the data.
            Any trade deltas are snapped to the exact timeframe,
            so that when candle data gets updated,
            we always update only full candles and
            never return trade data that would break the timeframe in a middle.

            Default to one minute candles.

        :param data_retention_time:
            Discard entries older than this to avoid
            filling the RAM.

        """

        assert isinstance(reorg_mon, ReorganisationMonitor)

        for o in oracles.values():
            assert isinstance(o, BaseOracle)

        self.pairs = pairs
        self.oracles = oracles
        self.data_retention_time = data_retention_time
        self.reorg_mon = reorg_mon
        self.tqdm = tqdm
        self.timeframe = timeframe
        self.cycle = 1

        self.trades_df = pd.DataFrame()

        # Check that every pair has a exchange rate conversion oracle
        for p in self.pairs:
            assert p in self.oracles, f"Pair {p} lacks price oracle. Add oracles by pair identifiers."

    def get_block_number_of_last_trade(self) -> Optional[int]:
        """Get the last block number for which we have good data."""

        if len(self.trades_df) == 0:
            return None

        return self.trades_df.iloc[-1]["block_number"]

    def add_trades(self, trades: Iterable[Trade]):
        """Add trade to the ring buffer with support for fixing chain reorganisations.

        Transactions may hop between different blocks when the chain tip reorganises,
        getting a new timestamp. In this case, we update the

        .. note::

            It is safe to call this function multiple times for the same event.

        :return:
            True if the transaction hopped to a different block

        :raise ChainReorganisationDetected:
            If we have detected a block reorganisation
            during importing the data

        """
        data = []

        for evt in trades:
            assert isinstance(evt, Trade)
            self.reorg_mon.check_block_reorg(evt.block_number, evt.block_hash)
            data.append(asdict(evt))

        new_data = pd.DataFrame(data, columns=list(Trade.get_dataframe_columns().keys()))
        new_data.set_index("block_number", inplace=True, drop=False)
        self.trades_df = pd.concat([self.trades_df, new_data])

    def get_latest_trades(self, n: int, pair: Optional[PairId] = None) -> pd.DataFrame:
        """Returns the latest trades.

        These trades will be across all trading pairs we are monitoring.

        :param n:
            Number of trades to return

        :param pair:
            Optional pair to filter with

        :return:
            DataFrame containing n trades
        """
        if pair:
            df = self.trades_df.loc[self.trades_df["pair"] == pair]
        else:
            df = self.trades_df
        return df.tail(n)

    def truncate_reorganised_data(self, latest_good_block):
        """Discard data because of the chain reorg.

        :param latest_good_block:
            The last block that we cannot discard.
        """

        if len(self.trades_df) > 0:
            self.trades_df = self.trades_df.truncate(after=latest_good_block, copy=False)

    def check_reorganisations_and_purge(self) -> ChainReorganisationResolution:
        """Check if any of block data has changed

        :return:
            Start block since we need to read new data (inclusive).
        """
        reorg_resolution = self.reorg_mon.update_chain()
        self.truncate_reorganised_data(reorg_resolution.latest_block_with_good_data)
        return reorg_resolution

    def backfill_buffer(self, block_count: int, tqdm: Optional[Type[tqdm]] = None) -> TradeDelta:
        """Populate the backbuffer before starting real-time tracker.

        :param block_count:
            Number of blocks we need to fetch

        :param tqdm:
            A progress visualiser.

            Especially useful during the initial fetch, to show the user how
            long it takes time to fill the buffer.

            Must be `tqdm` context manager compatible.

        :return:
            Data loaded and filled to the work buffer.
        """
        start_block, end_block = self.reorg_mon.load_initial_data(block_count)
        trades = self.fetch_trades(start_block, end_block, tqdm)
        # On initial load, we do not care about reorgs
        return self.update_cycle(start_block, end_block, False, trades)

    def get_exchange_rate(self, pair: str) -> Decimal:
        """Get the current exchange rate for the pair.

        - There is no block number parameter; We always assume
          the data is filled in order, block by block,
          so the exchange rate is always the exchange rate of the current block.

        - Price oracles are updates per-block before
          trade feed for pairs, so price orocles
          can calculate TWAP or similar exchange rate
          for the block.
        """
        return self.oracles[pair].calculate_price()

    def find_first_included_block_in_candle(self, ts: pd.Timestamp) -> Optional[int]:
        """Find the first block number that contains data going into the candle.

        :param ts:
            Timestamp when the block can start (inclusive)

        :return:
            Block number.

            If there is no data, return None.
        """

        after_df = self.trades_df.loc[self.trades_df["timestamp"] >= ts]

        if len(after_df) > 0:
            return after_df.iloc[0]["block_number"]
        else:
            return None

    def update_cycle(self, start_block, end_block, reorg_detected, trades: Iterable[Trade]) -> TradeDelta:
        """Update the internal work buffer.

        - Adds the new trades to the work buffer

        - Updates the cycle number

        - Creates the snapshot of the new trades for the client

        :param start_block:
            Incoming new data, first block (inclusive)

        :param end_block:
            Incoming new data, last  block (inclusive)

        :param reorg_detected:
            Did we detect any chain reorganisations in this cycle

        :param trades:
            Iterable o new trades

        :return:
            Delta of new trades
        """

        self.add_trades(trades)

        # We need to snap any trade delta update to the edge of candle timeframe
        event_start_ts = self.reorg_mon.get_block_timestamp_as_pandas(start_block)
        data_start_ts = self.timeframe.round_timestamp_down(event_start_ts)
        data_start_block = self.find_first_included_block_in_candle(data_start_ts)

        # If we do not have enough data, then just naively use the given start block
        # and have half-finished candle
        snap_block = data_start_block or start_block

        if len(self.trades_df) > 0:
            exported_trades = self.trades_df.loc[snap_block:end_block+1]
        else:
            exported_trades = pd.DataFrame()

        start_ts = pd.Timestamp.utcfromtimestamp(self.reorg_mon.get_block_timestamp(snap_block)).tz_localize(None)
        end_ts = pd.Timestamp.utcfromtimestamp(self.reorg_mon.get_block_timestamp(snap_block)).tz_localize(None)

        res = TradeDelta(
            self.cycle,
            snap_block,
            start_block,
            end_block,
            start_ts,
            end_ts,
            reorg_detected,
            exported_trades,
        )

        self.cycle += 1

        return res

    def perform_duty_cycle(self) -> TradeDelta:
        """Update the candle data

        1. Check for block reorganisations

        2. Read new data

        3. Process and index data to candles
        """
        reorg_resolution = self.check_reorganisations_and_purge()
        start_block = reorg_resolution.latest_block_with_good_data + 1
        end_block = self.reorg_mon.get_last_block_read()

        if start_block > end_block:
            # This situation can happen when the lsat block in the chain has reorganised,
            # in such case just read the last block again
            start_block = end_block

        trades = self.fetch_trades(start_block, end_block)
        return self.update_cycle(start_block, end_block, reorg_resolution.reorg_detected, trades)

    def restore(self, df: pd.DataFrame):
        """Restore data from the previous disk save."""
        self.trades_df = df

    @abstractmethod
    def fetch_trades(self,
                     start_block: int,
                     end_block: Optional[int],
                     tqdm: Optional[Type[tqdm]] = None
                     ) -> Iterable[Trade]:
        """Read data from the chain.

        Add any new trades using :py:meth:`add_trades`.

        :param start_block:
            Start reading from this block (inclusive)

        :param end_block:
            End at this block (inclusive)

        :param tqdm:
            Optional progress bar displayer gadget.

        :raise ChainReorganisationDetected:
            If blockchain detects minor reorganisation during the data ignestion

        :return:
            TradeDelta instance that contains all new trades since start_block.
        """
